module Model {
	// MineSweeper で発生したイベントのコールバックに使う。
	export interface OnMineSweeperEventListener {
		// ステージが生成された時
		onCreate(board: MineSweeper): void;
		// 通常のセルが開かれた時。地雷セルの時は呼ばれない。
		onOpen(cell: Cell): void;
		// 地雷セルが開かれた時。通常セルの時は呼ばれない。
		onDead(cell: Cell): void;
		// ゲームクリアした時。
		onClear(duration: number): void;
	}

	// 盤面のセル1個を表す。
	export class Cell {
		private open = false;

		// コンストラクタ。4つのフィールドに渡された値を初期値としてセットする。
		constructor(private board: MineSweeper, public x: number, public y: number, public bomb: bool) {
		}

		// 周囲8セルに爆弾がいくつあるかを返す。
		get neighbors(): number {
			return this.board.getNeighborsBomb(this.x, this.y);
		}

		// このセルが既に開かれているかを返す。
		get opened(): bool {
			return this.open;
		}

		set opened(val: bool) {
			this.open = val;
		}
	}

	// マインスイーパのゲーム制御のクラス。描画は行わない。
	export class MineSweeper {
		// 盤面の状態
		private board: Cell[] = [];
		private start = new Date();

		// コンストラクタ。初期の盤面の生成を行う。
		constructor(private listener: OnMineSweeperEventListener, public width?: number = 10, public height?: number = 10, public threshold?: number = 0.15) {
			for(var w = 0; w < this.width; w++) {
				for(var h = 0; h < this.height; h++) {
					var cell = new Cell(this, w, h, Math.random() < this.threshold);
					this.board.push(cell);
				}
			}

			this.listener.onCreate(this);
		}

		// 指定箇所のセルを取得する。
		getCell(x: number, y: number): Cell {
			var cells = this.board.filter(c => c.x == x && c.y == y);
			return cells[0];
		}

		// 指定箇所の周囲8セルを取得する。
		getNeighbors(x: number, y: number): Cell[] {
			var neighbors = this.board.filter(c => {
				return (c.x == x && c.y == y) == false && (x - 1 <= c.x && c.x <= x + 1) && (y - 1 <= c.y && c.y <= y + 1);
			});
			return neighbors;
		}

		/*
		 * 指定箇所のセルを開く。再帰的に開ける場合はそこも開く。
		 * onClearが複数回呼ばれると都合が悪いので openについて root がtrueかどうかで呼ぶかどうかを判断する。
		 */
		open(x: number, y: number, root?: bool = true): void {
			var cell = this.getCell(x, y);
			if(cell.opened) {
				return;
			}
			cell.opened = true;
			if(cell.bomb) {
				this.listener.onDead(cell);
			} else {
				this.listener.onOpen(cell);
				// 周囲8マス全てが地雷じゃない場合再帰的に開いていく。
				if(cell.neighbors == 0) {
					// 周囲8マスのセルを開ける。
					this.getNeighbors(x, y).forEach(c => this.open(c.x, c.y, false));
				}
			}
			if(root && this.cleared) {
				this.listener.onClear(new Date().getTime() - this.start.getTime());
			}
		}

		// 指定箇所が爆弾かどうかを調べる。
		isBomb(x: number, y: number): bool {
			return this.getCell(x, y).bomb;
		}

		// クリア済の状態かどうかを調べる。
		get cleared(): bool {
			// 閉じてる非地雷セルがあったら終わってない。
			var closedCells = this.board.filter(c => !c.opened && !c.bomb);
			var closedCellExists = closedCells.length != 0;

			// 開いてる地雷セルがあったらクリアではない。
			var openedBombs = this.board.filter(c => c.opened && c.bomb);
			var openedBombsExists = openedBombs.length != 0;

			return !closedCellExists && !openedBombsExists;
		}

		// 指定箇所の周囲に爆弾がいくつあるかを調べる。
		getNeighborsBomb(x: number, y: number): number {
			var bomb = 0;
			this.getNeighbors(x, y).forEach(c => bomb += c.bomb ? 1 : 0);
			return bomb;
		}

		get cells(): Cell[] {
			return this.board;
		}
	}
}